#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <assert.h>
#include <netinet/ip_icmp.h>
#include <sys/time.h>
#include "hash.h"
#include "debug.h"
#include "aprobe.h"

struct context * ctx = NULL;

#ifdef DEBUG
char * otv(struct timeval* tv)
{
	static char buffer[100];
	snprintf(buffer, 100, "%d.%d\n", tv->tv_sec, tv->tv_usec);
	return(buffer);
}
struct timeval dtv[100];
#endif

void help(void)
{
	printf("aprobe (c) 2005 Mateusz Golicz\n"
	       "usage:\n"
	       "  aprobe [-i <interval>] [-s <length>] [-f <out-filename>] <configuration file name>\n\n"
	       "options:\n"
	       "  -s <length>     Append bogus data to each packet\n"
	       "  -f <file>       File name to make reports to\n"
	       "\n"
	       "Configuration file syntax: \n"
	       "policy { \n"
	       "	heartbeat: 0.25;\n"
	       "	bucket {\n"
	       "		span: 10;\n"
	       "		limit: 3;\n"
	       "		delay-threshold: 100;\n"
	       "	};\n"
	       "	bucket {\n"
	       "		span: 300;\n"
	       "		limit: 20;\n"
	       "		delay-threshold: 50;\n"
	       "	};\n"
	       "	host {\n"
	       "		label: core-leaf;\n"
	       "		ip: 217.173.160.1;\n"
	       "	};\n"
	       "}\n"
	       "Send me an USR1 to cause a report dump\n");
}


void report(int sig)
{
	struct entry *e;
	FILE *fh;
	
	signal(SIGALRM, SIG_IGN);
	
	if(ctx->filename)
	{
		if(!(fh = fopen(ctx->filename, "w")))
		{
			signal(SIGALRM, send_pings);
			return;
		}
	}
	else
		fh = stdout;
	
	fprintf(fh, "# %-20s %10s %10s %6s %6s %10s\n", 
		"Label", "Sent", "Rcvd", "Loss", "Availa", "Avg rtt");
	for(e = ctx->hosts; e; e = e->next)
	{
		fprintf(fh, "  %-20s %10d %10d %6.1f %6.1f %10.2f\n",
			e->label,
			e->sent,
			e->recv,
			e->sent ? (((double) e->sent - e->recv) * 100.0 / ((double) e->sent)) : 0,
			e->sent ? (((double) e->recv) * 100.0 / ((double) e->sent)) : 0,
			e->sum / ((double) e->accounted) / 1000);
		e->sent = e->recv = 0;
		e->sum = 0;
		e->accounted = 0;
		memset(e->tv, 0, sizeof(e->tv));
	}
	if(ctx->filename)
		fclose(fh);
	
	signal(SIGALRM, send_pings);
}

int main(int argc, char **argv)
{
	int c, size = 8, verbose = 0;
	struct itimerval itv;
	char *fname = NULL;
	double interval;
	setlinebuf(stdout);
		
	while((c = getopt(argc, argv, "i:?hs:vf:")) != EOF)
	{
		switch(c)
		{
			case 'h':
			case '?':
				help();
				return 1;
			case 'f':
				fname = strdup(optarg);
				break;
			case 'v':
				verbose = 1;
				break;
			case 's':
				size = atoi(optarg);
				if(size < 0)
				{
					fprintf(stderr, "Invalid size: %s\n", optarg);
					return 1;
				}
				break;
		}
	}
	
	if(!argv[optind])
	{
		help();
		return 1;
	}

	if(argv[optind + 1])
	{
		fprintf(stderr, "Unrecognized commandline arguments (starting from: %s)\n", argv[optind + 1]);
		return 1;
	}
	
	ctx = cfg_read(argv[optind]);
	if(!ctx)
		return 1;
	
	debug("[main] Finished reading configuration\n");
	
	ctx->filename = fname;
	ctx->size = size;
	ctx->verbose = verbose;
	
	if((ctx->skt = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0)
	{
		perror("socket");
		return 1;
	}
	fcntl(ctx->skt, F_SETFL, O_NONBLOCK);
	ctx->pid = getpid() & 0xffff;
	
	signal(SIGALRM, send_pings);
	signal(SIGUSR1, report);
	
	debug("[main] pps estimated at = %.2f\n", ctx->pps);
	ctx->pps = ceil(ctx->pps + 1);

	interval = 1 / ctx->pps;

        itv.it_interval.tv_sec = interval;
        itv.it_interval.tv_usec = ((int) (interval * 1000000)) % 1000000;
        itv.it_value.tv_sec = interval;
        itv.it_value.tv_usec = ((int) (interval * 1000000)) % 1000000;
        setitimer(ITIMER_REAL, &itv, NULL);
	
	debug("[main] timer interval = %.2f\n", interval);
	
	for(;;)
	{
		fd_set fds;
		FD_ZERO(&fds);
		FD_SET(ctx->skt, &fds);
/*		send_pings(0); */

#ifdef DEBUG
		gettimeofday(&dtv[0], NULL);
#endif
		if(select(ctx->skt + 1, &fds, NULL, NULL, NULL) < 0)
		{
			if(errno != EINTR)
				perror("select");
		}
#ifdef DEBUG
		gettimeofday(&dtv[1], NULL);
#endif
		
		if(FD_ISSET(ctx->skt, &fds))
			recv_pings(ctx);
		
	}
}
