#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include "hash.h"
#include "configfile.h"
#include "debug.h"
#include "aprobe.h"
#include "memory.h"

struct context * cfg_read(char *filename)
{
	struct context * r;
	cssctx * ctx = NULL;
	cssctx * w = NULL;
	char errors[ERRORS + 1];
	
	ctx = css_ParseFile(filename, errors, ERRORS);
	if(!ctx)
	{
		fprintf(stderr, "Error while parsing the configuration file: %s\n", errors);
		return NULL;
	}
	
	r = (struct context*) mem_alloc("Probing context storage", sizeof(struct context));
	
	r->policies = NULL;
	r->hosts = NULL;
	r->ip_hash = hash_create(HS_HOST);
	r->label_hash = hash_create(HS_HOST);
	r->freq = 1;
	
	while((w = css_FindChild(ctx, w, "policy")))
	{
		cssval *v;
		struct policy *p;
		cssctx *b = NULL;
		cssctx *h = NULL;
		
		p = (struct policy*) mem_alloc("Probing policy storage", sizeof(struct policy));
		p->freq = 0.2;
		
		if((v = css_FindValue(w, NULL, "heartbeat")))
		{
			char *eptr;
			p->freq = strtod(v->value, &eptr);
			if(*eptr)
			{
				fprintf(stderr, "Error: Unable to parse 'heartbeat' tag in 'policy' context: '%s'\n", v->value);
				goto err;
			}
		}
		
		debug("Policy: freq = %.2f\n", p->freq);		

		while((b = css_FindChild(w, b, "bucket")))
		{
			struct bucketinfo *bi;
			bi = mem_alloc("Policy bucket information storage", sizeof(struct bucketinfo));
			bi->next = p->buckets;
			p->buckets = bi;
			
			bi->limit = 10;
			bi->span = 300;
			bi->threshold = 25;
			bi->name = NULL;

			if((v = css_FindValue(b, NULL, "span")))
			{
				char *eptr;
				bi->span = strtod(v->value, &eptr);
				if(*eptr)
				{
					fprintf(stderr, "Error: Unable to parse 'span' tag in 'bucket' context: '%s'\n", v->value);
					goto err;
				}
			}

			if((v = css_FindValue(b, NULL, "delay-threshold")))
			{
				char *eptr;
				bi->threshold = strtod(v->value, &eptr);
				if(*eptr)
				{
					fprintf(stderr, "Error: Unable to parse 'threshold' tag in 'bucket' context: '%s'\n", v->value);
					goto err;
				}
			}

			if((v = css_FindValue(b, NULL, "limit")))
			{
				char *eptr;
				bi->limit = strtod(v->value, &eptr);
				if(*eptr)
				{
					fprintf(stderr, "Error: Unable to parse 'limit' tag in 'bucket' context: '%s'\n", v->value);
					goto err;
				}
				bi->limit *= p->freq * bi->span / 100; /* Limit is in percent */
				
			}			
			

			if((v = css_FindValue(b, NULL, "name")))
				bi->name = strdup(v->value);
			

			debug("Bucket: span = %.2f limit = %.2f threshold = %.2f name = %s\n", bi->span, bi->limit, bi->threshold, bi->name);
		}

		p->next = r->policies;
		r->policies = p;

		while((h = css_FindChild(w, h, "host")))
		{
			struct entry * ho;
			int la = 0, ip = 0;
			struct bucketinfo *bi;
			
			r->pps += p->freq;
			ho = (struct entry*) mem_alloc("Host storage", sizeof(struct entry));
			ho->policy = p;
			
			if((v = css_FindValue(h, NULL, "label")))
			{
				if(hash_get(r->label_hash, v->value))
				{
					fprintf(stderr, "Error: Duplicate host label ('%s')\n", v->value);
					goto err;
				}
				ho->label = v->value;
				la = 1;
			}

			if((v = css_FindValue(h, NULL, "ip")))
			{
				if(!inet_aton(v->value, &ho->ip.sin_addr))
				{
					fprintf(stderr, "Error: Invalid IP address ('%s')\n", v->value);
					goto err;
				}
				if(hash_get(r->ip_hash, inet_ntoa(ho->ip.sin_addr)))
				{
					fprintf(stderr, "Error: Duplicate host IP ('%s')\n", v->value);
					goto err;
				}
				ip = 1;
			}

			
			if(!(la && ip))
			{
				fprintf(stderr, "Error: Incomplete host definition (both label and IP required)\n");
				goto err;
			}
			
			debug("Host: label = %s, ip = %s\n", ho->label, inet_ntoa(ho->ip.sin_addr));
			hash_add(r->label_hash, ho->label, ho);
			hash_add(r->ip_hash, inet_ntoa(ho->ip.sin_addr), ho);
			
			/* Create buckets */
			for(bi = p->buckets; bi; bi = bi->next)
			{
				struct bucket *b;
				b = (struct bucket*) mem_alloc("Bucket", sizeof(struct bucket));
				b->parent = bi;
				b->bucket = bi->limit;
				b->next = ho->buckets;
				ho->buckets = b;
			}
			
			ho->next = r->hosts;
			r->hosts = ho;
		}	
	}

	if(!r->hosts)
	{
		fprintf(stderr, "Nothing to do!\n");
		goto err;
	}

	return(r);

err:
	return(NULL);	
}
